use arrayvec::ArrayVec;
use bitvec::prelude::*;
use ham_cats::packet::Packet;

use crate::{decoder::Bit, MAX_PACKET_LEN};

pub enum Status {
    Some,
    NotYet,
    Never,
}

pub struct PacketDecoder {
    bits: ArrayVec<f32, { MAX_PACKET_LEN * 8 }>,
    len: Option<usize>,
    never: bool,
}

impl PacketDecoder {
    pub fn new() -> Self {
        Self {
            bits: ArrayVec::new(),
            len: None,
            never: false,
        }
    }

    pub fn push(&mut self, bit: f32) {
        if self.never {
            return;
        }

        let _ = self.bits.try_push(bit);
    }

    pub fn decode<F: Fn(Packet<MAX_PACKET_LEN>)>(&mut self, handler: F) -> Status {
        if self.never {
            return Status::Never;
        }

        let len = match self.len() {
            Some(x) => x,
            None => return Status::NotYet,
        };

        if self.bits.len() < 16 + 8 * len {
            return Status::NotYet;
        }

        // try decode
        // currently only hard bits
        let hardened: BitVec<u8, Msb0> = self
            .bits
            .iter()
            .skip(16)
            .take(len * 8)
            .map(|x| x.bit())
            .collect();

        let mut buf = [0; MAX_PACKET_LEN];
        match Packet::fully_decode(hardened.as_raw_slice(), &mut buf) {
            Ok(x) => {
                handler(x);
                Status::Some
            }
            Err(_) => {
                self.never = true;
                Status::Never
            }
        }
    }

    fn len(&mut self) -> Option<usize> {
        if let Some(len) = self.len {
            return Some(len);
        }

        let soft_bits = self.bits.get(0..16)?;

        let mut upper: u8 = 0;
        for b in &soft_bits[0..8] {
            upper <<= 1;
            upper |= u8::from(b.bit());
        }

        let mut lower: u8 = 0;
        for b in &soft_bits[8..] {
            lower <<= 1;
            lower |= u8::from(b.bit());
        }

        // only 12 bits are used,
        // so we can 0 out the ones we know must be 0
        lower &= 0b00011111;

        let len = usize::from(u16::from_le_bytes([upper, lower]));
        self.len = Some(len);

        Some(len)
    }
}
